/*
 * bsp.c
 *
 * 3D BSP compiler
 * Format of input polygon list file:
 *
 *      %d          Number of vertices (list ends if this is 0)
 *      %d,%d,%d    X, Y, Z
 *      ...         Repeated until all vertices for the polygon are defined
 *
 * Format of BSP output file:
 *
 *      %d,%d,%d:%d ID, Front ID, Back ID, Number of vertices (list ends if
 *                  this is 0)
 *      %d,%d,%d    X, Y, Z
 *      ...         Repeated until all vertices for the polygon are defined
 *
 * Polygons must be convex no matter which orientation they are rotated into.
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "types.h"
#include "bsp.h"


/*
 * Polygon List
 *
 * The polygon list simply contains polygon definitions. When polygons are
 * split during the BSP construction process, the split parts are added as
 * new entries to this list.
 */

static POLY *poly_list;

/*
 * BSP Tree
 */

static BSP  *bsp_tree;


/*****************************************************************************
 Polygon List Handling

 Code to manage the polygon list.
*****************************************************************************/

static void KillPolyListRecursive(POLY *l)
{
    if (l == NULL)
        return;

    KillPolyListRecursive(l->next);
    free(l);
}

static void KillPolyList()
{
    POLY    *next;

    while (poly_list != NULL)
    {
        next = poly_list->next;
        free(poly_list->vertex);
        free(poly_list);
        poly_list = next;
    }
}

static BOOL NotTheSameVertex(VERTEX *a, VERTEX *b)
{
    if (a->x != b->x || a->y != b->y || a->z != b->z)
        return 1;   // return 1 if the vertices are not the same
    return 0;
}

/*
 * Removes any redundancies in the vertex list and returns the number of
 * resultant vertices. If 0 is returned, the polygon ought to be discarded
 */

static INT RemoveRedundantVertices(VERTEX *out, VERTEX *in, INT num_vertices)
{
    INT i, j, out_vertices;

    if (num_vertices < 3)
        return 0;

    out_vertices = 0;

    for (i = 0; i < num_vertices; )
    {
        for (j = i; j < num_vertices; j++)
        {
            if (NotTheSameVertex(&in[j], &in[i]))   // scan to next different vertex
                break;
        }

        if (j < num_vertices)   // we have not exceeded the limits of the list
            out[out_vertices++] = in[i];
        else                    // check to see if last vertex == start vertex
        {
            if (NotTheSameVertex(&in[num_vertices - 1], &in[0]))  // they're different
                out[out_vertices++] = in[num_vertices - 1];
        }

        i = j;
    }

    return (out_vertices < 3) ? 0 : out_vertices;
}

static BOOL AddPolyToList(POLY **plist, POLY *p)
{
    POLY    *new;

    if ((new = malloc(sizeof(POLY))) == NULL)
        return 1;

    if ((new->vertex = malloc(sizeof(VERTEX) * p->num_vertices)) == NULL)
    {
        free(new);
        return 1;
    }

    new->num_vertices = RemoveRedundantVertices(new->vertex, p->vertex, p->num_vertices);
    new->valid = !!new->num_vertices;   // only valid if num_vertices != 0

    if (!new->valid || !p->valid)       // not a valid polygon?
    {
        free(new->vertex);
        free(new);
        return 0;
    }

    new->next = *plist;
    *plist = new;

    return 0;
}


/*****************************************************************************
 BSP Tree Code

 BSP construction, output, and deletion.
*****************************************************************************/

/*
 * DotProduct(): Performs a dot product between one vector and another which
 * is derived from a vertex
 */

static float DotProduct(VECTOR *u, VERTEX *v)
{
    return (u->x * v->x + u->y * v->y + u->z * v->z);
}

/*
 * MakePlaneOutOfPoly(): Creates a plane on which the specified polygon lies.
 * If the plane can't be found in which the polygon lies (when a unit normal
 * cannot be calculated), 1 is returned (the polygon should be discarded)
 */

static BOOL MakePlaneOutOfPoly(PLANE *plane, POLY *poly)
{
    VECTOR  side1, side2, c;    // 2 side vectors and a cross product vector
    float   len;
    INT     i, j, k;

    /*
     * Find a normal vector.
     *
     * Performing a cross product will find a normal which points towards the
     * polygon, The polygon splitter code (which was derived from a 3D polygon
     * clipper) assumes that the normal points TOWARDS the polygon.
     *
     * The first 2 polygon sides are converted into vectors. If a (0,0,0)
     * vector is returned, we keep looping through the different edges to find
     * a better normal. If none is found, tough luck.
     */

    i = 0;    
    do
    {
        /*
         * j points to the next vertex after i, and k to the next one after
         * that
         */

        j = ((i + 1) >= poly->num_vertices) ? ((i + 1) - poly->num_vertices) : (i + 1);
        k = ((i + 2) >= poly->num_vertices) ? ((i + 2) - poly->num_vertices) : (i + 2);

        side1.x = poly->vertex[j].x - poly->vertex[i].x;
        side1.y = poly->vertex[j].y - poly->vertex[i].y;
        side1.z = poly->vertex[j].z - poly->vertex[i].z;

        side2.x = poly->vertex[k].x - poly->vertex[j].x;
        side2.y = poly->vertex[k].y - poly->vertex[j].y;
        side2.z = poly->vertex[k].z - poly->vertex[j].z;

        c.x = side1.y * side2.z - side1.z * side2.y;
        c.y = side1.z * side2.x - side1.x * side2.z;
        c.z = side1.x * side2.y - side1.y * side2.x;

        i++;
        if (i == poly->num_vertices)
            i = 0;
    }
    while (c.x == 0 && c.y == 0 && c.z == 0 && i != 0);

    if (c.x == 0 && c.y == 0 && c.z == 0)
    {
        fprintf(stderr, "bsp: warning: polygon with invalid normal vector (0,0,0) discarded\n");
#if 0
        for (i = 0; i < poly->num_vertices; i++)
            printf("\t(%f,%f,%f)\n", poly->vertex[i].x, poly->vertex[i].y, poly->vertex[i].z);
        puts("");
#endif
        return 1;
    }

    /*
     * Turn the vector into a unit vector (length of 1)
     */

    len = sqrt(c.x * c.x + c.y * c.y + c.z * c.z);

    plane->unit_normal.x = c.x / len;
    plane->unit_normal.y = c.y / len;
    plane->unit_normal.z = c.z / len;


    /*
     * Calculate the distance from the origin to the plane along the unit
     * normal by performing a dot product of the normal vector and a point on
     * the plane (any point in the polygon)
     */

    plane->dist = DotProduct(&(plane->unit_normal), &(poly->vertex[0]));
    return 0;
}

/*
 * The output vertex list for the inside polygon may be the same as the input
 * vertex list
 */

void SplitPolyWithPlane(VERTEX *outlist_inside, VERTEX *outlist_outside, INT *num_vertices_inside, INT *num_vertices_outside, VERTEX *inlist, PLANE *plane, INT num_vertices)
{  
    INT         i, current_in, next_in, next_i, out_vertices_in, out_vertices_out;
    float       current_dot, next_dot, scale;
    VERTEX      *in;

    in = inlist;

    /*
     * Find distance of current vertex
     */

    current_dot = DotProduct(&(plane->unit_normal), in);
    current_in = (current_dot <= plane->dist);  // 1 if inside plane

    for (i = out_vertices_in = out_vertices_out = 0; i < num_vertices; i++)
    {
        if (current_in)
        {
            *outlist_inside++ = *in;
            out_vertices_in++;
        }
        else
        {
            *outlist_outside++ = *in;
            out_vertices_out++;
        }

        next_i = i + 1;             // next vertex in polygon
        if (next_i >= num_vertices)
            next_i = 0;

        next_dot = DotProduct(&(plane->unit_normal), &(inlist[next_i]));
        next_in = (next_dot <= plane->dist);    // 1 if inside plane

        if (next_in != current_in)  // there is an intersection somewhere!
        {
            scale = (plane->dist - current_dot) / (next_dot - current_dot);

            outlist_inside->x = in->x + (inlist[next_i].x - in->x) * scale;
            outlist_inside->y = in->y + (inlist[next_i].y - in->y) * scale;
            outlist_inside->z = in->z + (inlist[next_i].z - in->z) * scale;

            outlist_outside->x = outlist_inside->x;
            outlist_outside->y = outlist_inside->y;
            outlist_outside->z = outlist_inside->z;

            outlist_inside++;
            out_vertices_in++;

            outlist_outside++;
            out_vertices_out++;
        }

        current_dot = next_dot;
        current_in = next_in;
        in++;        
    }

    *num_vertices_inside = out_vertices_in;
    *num_vertices_outside = out_vertices_out;
}

static BOOL ClassifyPoly(POLY *poly, PLANE *plane, POLY **frontlist, POLY **backlist)
{
    POLY    inside, outside;

    /*
     * Allocate memory for the 2 polygons (one front, one back)
     */

    if ((inside.vertex = malloc(sizeof(VERTEX) * (poly->num_vertices + 1))) == NULL)
        return 1;

    if ((outside.vertex = malloc(sizeof(VERTEX) * (poly->num_vertices + 1))) == NULL)
    {
        free(inside.vertex);
        return 1;
    }

    inside.valid = outside.valid = 1;

    /*
     * Perform a split
     */

    SplitPolyWithPlane(inside.vertex, outside.vertex, &(inside.num_vertices), &(outside.num_vertices), poly->vertex, plane, poly->num_vertices);

    /*
     * There are now 3 possible scenarios:
     *
     *      1. The polygon was entirely inside the clip plane. We only have to
     *         add this polygon to the front list and we're done.
     *      2. The polygon was entirely outside the clip plane. We only have
     *         to add this polygon to the back list and we're done.
     *      3. The polygon was split and we have to add both parts to the
     *         proper list.
     */

    if (inside.num_vertices)
    {
        if (inside.num_vertices < 3)    // something is seriously wrong
            fprintf(stderr, "bsp: warning: an inside polygon was made with less than 3 vertices\n");

        if (AddPolyToList(frontlist, &inside))
        {
            free(inside.vertex);
            free(outside.vertex);
            return 1;
        }
    }

    if (outside.num_vertices)
    {
        if (outside.num_vertices < 3)   // something is seriously wrong
            fprintf(stderr, "bsp: warning: an outside polygon was made with less than 3 vertices\n");

        if (AddPolyToList(backlist, &outside))
        {
            free(inside.vertex);
            free(outside.vertex);
            return 1;
        }
    }

    free(inside.vertex);
    free(outside.vertex);
    return 0;
}

/*
 * Returns non-zero if a memory error occured. If *root is set to NULL, no
 * polygons were suitable to be a root (the current node should be set
 * to NULL.)
 */

static BOOL ChooseRootPoly(POLY **root, PLANE *best_plane, POLY *poly_list)
{
    PLANE   plane;
    POLY    *current, *candidate, *best_candidate, *front, *back;
    INT     num_splits, best_num_splits;

    best_candidate = NULL;
    best_num_splits = 0;

    /*
     * For each polygon, make a plane and then try to split all the other
     * polygons with it to see how many will need to be split
     */

    candidate = poly_list;
    while (candidate != NULL)
    {
        if (MakePlaneOutOfPoly(&plane, candidate))
        {
            candidate->valid = 0;   // could not make plane: invalid polygon
            candidate = candidate->next;
            continue;
        }

        /*
         * Try to split everything else
         */

        num_splits = 0;
        current = poly_list;
        while (current != NULL)
        {
            if (current == candidate || !current->valid)    // skip the candidate root poly and invalid polys
            {
                current = current->next;
                continue;
            }

            front = back = NULL;
            if (ClassifyPoly(current, &plane, &front, &back))
                return 1;

            if (front != NULL && back != NULL)
                num_splits++;

            if (front != NULL)  // discard this stuff, it's no longer needed
            {
                free(front->vertex);
                free(front);
            }
            if (back != NULL)
            {
                free(back->vertex);
                free(back);
            }

            current = current->next;
        }

        /*
         * Check to see if this was the best case thus far
         */

        if (best_candidate == NULL || num_splits < best_num_splits)
        {
            best_candidate = candidate;
            best_num_splits = num_splits;
            memcpy(best_plane, &plane, sizeof(PLANE));
        }

        candidate = candidate->next;
    }

    *root = best_candidate;
    return 0;
}

/*
 * NOTE: If non-zero is returned, the caller must destroy what was built of
 * the BSP tree and the polygon list
 */

static BOOL BuildBSPTree(BSP **tree, POLY *poly_list)
{
    PLANE       plane;
    POLY        *frontlist, *backlist, *root, *current;
    INT         error_code, i;

    /*
     * Choose a root polygon (also creates the splitting plane for the
     * polygon)
     */

    *tree = NULL;

    if (ChooseRootPoly(&root, &plane, poly_list))
        return 1;       // memory error!
    if (root == NULL)   // no good polygon was found
        return 0;

    /*
     * Categorize all polygons (except the root and invalid polygons) as
     * either in front or behind the root polygon and add them to the
     * appropriate list. This is done by trying to split them and then
     * checking to see if a split was actually performed or if the polygon
     * lies completely on one side of the splitter
     */

    frontlist = backlist = NULL;

    current = poly_list;
    while (current != NULL)
    {
        if (current == root || !current->valid) // skip the root poly and invalid polys
        {
            current = current->next;
            continue;
        }

        /*
         * Classify the polygon as front or back (it will be split if
         * necessary)
         */

        if (ClassifyPoly(current, &plane, &frontlist, &backlist))
        {
            KillPolyListRecursive(frontlist);
            KillPolyListRecursive(backlist);
            return 1;
        }

        current = current->next;
    }

    /*
     * Create a BSP tree node out of the root poly and add the front and back
     * lists to it
     */

    error_code = 0;
    if ((*tree = malloc(sizeof(BSP))) == NULL)
    {
        error_code = 1;
        goto END;
    }

    if (((*tree)->vertex = malloc(sizeof(VERTEX) * root->num_vertices)) == NULL)
    {
        error_code = 1;
        goto END;
    }

    (*tree)->num_vertices = root->num_vertices;

    for (i = 0; i < root->num_vertices; i++)
        (*tree)->vertex[i] = root->vertex[i];

    (*tree)->front = NULL;
    (*tree)->back = NULL;

    if (frontlist != NULL)
    {
        if (BuildBSPTree(&((*tree)->front), frontlist))
        {
            error_code = 1;
            goto END;
        }
    }

    if (backlist != NULL)
    {
        if (BuildBSPTree(&((*tree)->back), backlist))
        {
            error_code = 1;
            goto END;
        }
    }

END:
    KillPolyListRecursive(frontlist);
    KillPolyListRecursive(backlist);
        
    return error_code;
}

static void KillBSPTree(BSP *b)
{
    if (b == NULL)
        return;

    KillBSPTree(b->front);
    KillBSPTree(b->back);

    free(b);
}

static void OutputBSPTree(FILE *fp, BSP *b)
{
    INT i;

    if (b == NULL)
        return;

    /*
     * It is important that the parent gets outputted before the children so
     * that the renderer can properly re-construct the tree
     */

    fprintf(fp, "%d,%d,%d:%d\n", (INT) b, (INT) b->front, (INT) b->back, b->num_vertices);
    for (i = 0; i < b->num_vertices; i++)
        fprintf(fp, "%f,%f,%f\n", b->vertex[i].x, b->vertex[i].y, b->vertex[i].z);

    OutputBSPTree(fp, b->front);
    OutputBSPTree(fp, b->back);
}


/*****************************************************************************
 Main Program Code

 Usage is:

    bsp <file> [outfile]

 Where <file> is a polygon list. The output is written to stdout unless an
 output file is explicitly specified.
*****************************************************************************/

static BOOL LoadInput(char *file)
{
    FILE    *fp;
    POLY    *p;    
    INT     i, j;

    if ((fp = fopen(file, "r")) == NULL)
    {
        fprintf(stderr, "bsp: unable to open input file: %s\n", file);
        return 1;
    }

    while (1)
    {
        fscanf(fp, "%d", &i);   // get the number of vertices
        if (!i)                 // 0? this indicates the file is done
            break;

        if ((p = malloc(sizeof(POLY))) == NULL)
        {
            fprintf(stderr, "bsp: ran out of memory while trying to load %s\n", file);
            fclose(fp);
            KillPolyList();
            return 1;
        }

        /*
         * Allocate the required number of vertices + 1 (when clipping a
         * convex [in all orientations] polygon against a plane, I've found
         * that, at most, 1 extra vertex may be needed)
         */

        if ((p->vertex = malloc(sizeof(VERTEX) * (i + 1))) == NULL)
        {
            fprintf(stderr, "bsp: ran out of memory while trying to load %s\n", file);
            fclose(fp);
            free(p);
            KillPolyList();
            return 1;
        }

        p->num_vertices = i;
        for (j = 0; j < i; j++)
            fscanf(fp, "%f,%f,%f", &(p->vertex[j].x), &(p->vertex[j].y), &(p->vertex[j].z));

        p->valid = 1;

        /*
         * Insert into beginning of polygon list (for simplicity)
         */

        p->next = poly_list;
        poly_list = p;
    }

    fclose(fp);
    return 0;
}
    
/*
 * int main(int argc, char **argv);
 *
 * Program entry point.
 */

int main(int argc, char **argv)
{
    FILE    *out_fp;

    /*
     * Start up (open output file, load input file)
     */

    if (argc <= 1)
    {
        puts("bsp: Usage: bsp <infile> [outfile]");
        exit(0);
    }

    poly_list = NULL;
    bsp_tree = NULL;

    if (argc > 2)
    {
        if ((out_fp = fopen(argv[2], "w")) == NULL)
        {
            fprintf(stderr, "bsp: unable to open output file: %s\n", argv[2]);
            exit(1);
        }
    }
    else
        out_fp = stdout;

    if (LoadInput(argv[1]))
    {
        if (out_fp != stdout)
            fclose(out_fp);
        exit(1);
    }

    /*
     * Build the BSP Tree
     */

    if (BuildBSPTree(&bsp_tree, poly_list))
    {
        fprintf(stderr, "bsp: ran out of memory while trying to build BSP tree\n");
        if (out_fp != stdout)
            fclose(out_fp);
        KillPolyList();
        KillBSPTree(bsp_tree);
        exit(1);
    }

    OutputBSPTree(out_fp, bsp_tree);
    fprintf(out_fp, "0,0,0:0\n");   // end of BSP node list

    /*
     * Quit
     */

    KillPolyList();
    KillBSPTree(bsp_tree);
    if (out_fp != stdout)
        fclose(out_fp);

    return 0;
}

