#define FOV                 90.0
#define NEAR_CLIP_DISTANCE  0
#define FAR_CLIP_DISTANCE   5000

/*
 * r.c
 *
 * BSP renderer
 */

#include <conio.h>
#include <dpmi.h>
#include <go32.h>
#include <math.h>
#include <pc.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/farptr.h>
#include <time.h>
#include "types.h"
#include "angle.h"
#include "bsp.h"
#include "keyb.h"
#include "poly.h"
#include "scrshot.h"
#include "vesalib.h"
#include "vector.h"


/*
 * Node List
 */

static NODE *node_list;

/*
 * BSP Tree
 */

static BSP  *bsp_tree;

/*
 * Vertex Buffers
 *
 * Capable of holding the largest possible vertex list.
 */

static VERTEX3D *vlist_a, *vlist_b;
static VERTEX2D *vlist_2d;

/*
 * Projection and FOV Stuff
 */

static float    dphys;
static float    fov_y;

/*
 * View Position and Orientation
 */

static float    x = 0, y = 0, z = -1000;
static float    yaw = 0.0, pitch = 0.0, roll = 0.0;

/*
 * Video Data
 */

static INT32            xres;
static INT32            yres;
static UINT8            *vidbuf;
static VESALIB_LFBINFO  lfb;


/*****************************************************************************
 Clipping
*****************************************************************************/

#define VERT2VEC(v, p_vert) { v.x = p_vert->x; v.y = p_vert->y; v.z = p_vert->z; }

static INT32 ClipPolyToPlane(VERTEX3D *outlist, VERTEX3D *inlist, PLANE *plane, INT32 num_vertices)
{  
    VECTOR3D    v;
    INT32       i, current_in, next_in, next_i, out_vertices;
    float       current_dot, next_dot, scale;
    VERTEX3D    *out, *in;

    out = outlist;  // for easy access
    in = inlist;

    // find distance of current vertex

    VERT2VEC(v, in);
    current_dot = Vec3D_RawDotProduct(&(plane->unit_normal), &v);
    current_in = (current_dot <= plane->dist);  // 1 if inside plane

    for (i = out_vertices = 0; i < num_vertices; i++)
    {
        if (current_in) // if the current vertex is inside, keep it
        {
            *out++ = *in;
            out_vertices++;
        }

        next_i = i + 1;             // next vertex in polygon
        if (next_i >= num_vertices)
            next_i = 0;

        VERT2VEC(v, (&inlist[next_i]));
        next_dot = Vec3D_RawDotProduct(&(plane->unit_normal), &v);
        next_in = (next_dot <= plane->dist);    // 1 if inside plane

        if (next_in != current_in)  // there is an intersection somewhere!
        {
            scale = (plane->dist - current_dot) / (next_dot - current_dot);

            out->x = in->x + (inlist[next_i].x - in->x) * scale;
            out->y = in->y + (inlist[next_i].y - in->y) * scale;
            out->z = in->z + (inlist[next_i].z - in->z) * scale;

            out++;
            out_vertices++;
        }

        current_dot = next_dot;
        current_in = next_in;
        in++;        
    }

    return out_vertices;
}

/*
 * The clipped vertices end up in the input vertex array, v. A scratch buffer
 * is used during the process
 */

static INT Clip(VERTEX3D *v, VERTEX3D *scratch, INT num_vertices)
{
    PLANE       frustum[6];
    VECTOR3D    view;

    view.x = view.y = 0;
    view.z = 0;

    /*
     * How Frustum Planes are Calculated:
     * ----------------------------------
     *       __--
     * eye --
     *       --
     *         --
     *
     * 1. Calculate a unit vector parallel to the plane using sin() and cos().
     * 2. Rotate this inwards (or is it outwards?) 90 degrees to create a
     *    normal suitable for clipping.
     * 3. Calculate the distance.
     * 4. Use this plane to clip :)
     *
     * NOTE: Clipping in world space is the same except that the unit normals
     * must be rotated so that the frustum is pointing where the camera is.
     * Some code like this will be required:
     *
     *      Vec3D_Rotate(&(frustum[1].unit_normal), 0, yaw, 0);
     */

    frustum[0].unit_normal.x = -sin(DEG2RAD(FOV / 2.0));
    frustum[0].unit_normal.y = 0;
    frustum[0].unit_normal.z = cos(DEG2RAD(FOV / 2.0));
    Vec3D_Rotate(&(frustum[0].unit_normal), 0, 90, 0);
    frustum[0].dist = Vec3D_RawDotProduct(&view, &(frustum[0].unit_normal));

    frustum[1].unit_normal.x = sin(DEG2RAD(FOV / 2.0));
    frustum[1].unit_normal.y = 0;
    frustum[1].unit_normal.z = cos(DEG2RAD(FOV / 2.0));
    Vec3D_Rotate(&(frustum[1].unit_normal), 0, -90, 0);
    frustum[1].dist = Vec3D_RawDotProduct(&view, &(frustum[1].unit_normal));

    frustum[2].unit_normal.x = 0;
    frustum[2].unit_normal.y = -sin(DEG2RAD(fov_y / 2.0));
    frustum[2].unit_normal.z = cos(DEG2RAD(fov_y / 2.0));
    Vec3D_Rotate(&(frustum[2].unit_normal), -90, 0, 0);
    frustum[2].dist = Vec3D_RawDotProduct(&view, &(frustum[2].unit_normal));

    frustum[3].unit_normal.x = 0;
    frustum[3].unit_normal.y = sin(DEG2RAD(fov_y / 2.0));
    frustum[3].unit_normal.z = cos(DEG2RAD(fov_y / 2.0));
    Vec3D_Rotate(&(frustum[3].unit_normal), 90, 0, 0);
    frustum[3].dist = Vec3D_RawDotProduct(&view, &(frustum[3].unit_normal));

    /*
     * Calculate far and near clip planes
     *
     * NOTE: The vector of the near clip plane points TOWARDS the eye.
     * Therfore, the distance must be negated because the distance is the
     * distance along the unit normal to the plane. If the vector is facing
     * us, a distance of X is actually X units behind us. Moving backwards
     * along the vector is what we need to do: -X.
     */

    frustum[4].unit_normal.x = 0;
    frustum[4].unit_normal.y = 0;
    frustum[4].unit_normal.z = -1;
    frustum[4].dist = -NEAR_CLIP_DISTANCE;

    frustum[5].unit_normal.x = 0;
    frustum[5].unit_normal.y = 0;
    frustum[5].unit_normal.z = 1;
    frustum[5].dist = FAR_CLIP_DISTANCE;

    /*
     * Do clip
     */

    num_vertices = ClipPolyToPlane(scratch, v, &frustum[0], num_vertices);
    num_vertices = ClipPolyToPlane(v, scratch, &frustum[1], num_vertices);
    num_vertices = ClipPolyToPlane(scratch, v, &frustum[2], num_vertices);
    num_vertices = ClipPolyToPlane(v, scratch, &frustum[3], num_vertices);
    num_vertices = ClipPolyToPlane(scratch, v, &frustum[4], num_vertices);
    num_vertices = ClipPolyToPlane(v, scratch, &frustum[5], num_vertices);

    return num_vertices;
}


/*****************************************************************************
 World Rendering
*****************************************************************************/

static void PerspectiveProject(VERTEX2D *out, VERTEX3D *in, INT32 num_vertices, float xo, float yo, float zo)
{
    INT32  i, xs, ys;
    float   x, y;

    xs = xres / 2;
    ys = yres / 2;

    for (i = 0; i < num_vertices; i++)
    {
        x = (in[i].x + xo) * dphys / (in[i].z + zo);
        y = (in[i].y + yo) * dphys / (in[i].z + zo);

        out[i].x = x + xs;
        out[i].y = y + ys;
    }
}

static void CameraTransform(VERTEX3D *out, VERTEX3D *in, INT32 num_v)
{
    INT32   i;
    float   cx, cy, cz;

    for (i = 0; i < num_v; i++)
    {
        cx = in[i].x - x;
        cy = in[i].y - y;
        cz = in[i].z - z;

        /*
         * Yaw 
         */

        out[i].x = cz * sin(DEG2RAD(yaw)) + cx * cos(DEG2RAD(yaw));
        out[i].y = cy;
        out[i].z = cz * cos(DEG2RAD(yaw)) - cx * sin(DEG2RAD(yaw));

        /*
         * Pitch
         */

        cx = out[i].x;
        cy = out[i].y * cos(DEG2RAD(pitch)) - out[i].z * sin(DEG2RAD(pitch));
        cz = out[i].y * sin(DEG2RAD(pitch)) + out[i].z * cos(DEG2RAD(pitch));

        /*
         * Roll
         */

        out[i].x = cy * sin(DEG2RAD(roll)) + cx * cos(DEG2RAD(roll));
        out[i].y = cy * cos(DEG2RAD(roll)) - cx * sin(DEG2RAD(roll));
        out[i].z = cz;
    }
}

static void DrawPoly(BSP *b)
{
    INT i;

    CameraTransform(vlist_a, b->node->vertex, b->node->num_vertices);
    i = Clip(vlist_a, vlist_b, b->node->num_vertices);
    PerspectiveProject(vlist_2d, vlist_a, i, 0, 0, 0);

    Polygon_DrawSolid(0, 0, b->node->color, xres, vidbuf, vlist_2d, i);
}

#if 0
static BOOL PolyIsFacingUs(BSP *b)
{
    VECTOR3D    side1, side2, c, view;
    float       dot;
    INT         i, j, k;

    /*
     * Find a normal vector to the polygon
     */

    i = 0;    
    do
    {
        // j points to the next vertex after i, and k to the next one after
        // that

        j = ((i + 1) >= b->node->num_vertices) ? ((i + 1) - b->node->num_vertices) : (i + 1);
        k = ((i + 2) >= b->node->num_vertices) ? ((i + 2) - b->node->num_vertices) : (i + 2);

        side1.x = b->node->vertex[j].x - b->node->vertex[i].x;
        side1.y = b->node->vertex[j].y - b->node->vertex[i].y;
        side1.z = b->node->vertex[j].z - b->node->vertex[i].z;

        side2.x = b->node->vertex[k].x - b->node->vertex[j].x;
        side2.y = b->node->vertex[k].y - b->node->vertex[j].y;
        side2.z = b->node->vertex[k].z - b->node->vertex[j].z;

        c.x = side1.y * side2.z - side1.z * side2.y;
        c.y = side1.z * side2.x - side1.x * side2.z;
        c.z = side1.x * side2.y - side1.y * side2.x;

        i++;
        if (i == b->node->num_vertices)
            i = 0;
    }
    while (c.x == 0 && c.y == 0 && c.z == 0 && i != 0);

    /*
     * Reverse the direction of the polygon's normal because the cross
     * product actually produces a vector which points away from the visible
     * side
     */

    c.x = -c.x;
    c.y = -c.y;
    c.z = -c.z;

    view.x = b->node->vertex[0].x - x;
    view.y = b->node->vertex[0].y - y;
    view.z = b->node->vertex[0].z - z;

    dot = Vec3D_RawDotProduct(&c, &view);
    return (dot <= 0) ? 1 : 0;
}
#endif
//#if 0
static BOOL PolyIsFacingUs(BSP *b)
{
    VECTOR3D    side1, side2, c, view;
    float       dot;

    /*
     * Find a normal vector to the polygon.
     *
     * The first 2 polygon sides are converted into vectors.
     */

    side1.x = b->node->vertex[1].x - b->node->vertex[0].x;
    side1.y = b->node->vertex[1].y - b->node->vertex[0].y;
    side1.z = b->node->vertex[1].z - b->node->vertex[0].z;

    side2.x = b->node->vertex[2].x - b->node->vertex[1].x;
    side2.y = b->node->vertex[2].y - b->node->vertex[1].y;
    side2.z = b->node->vertex[2].z - b->node->vertex[1].z;

    c.x = side1.y * side2.z - side1.z * side2.y;
    c.y = side1.z * side2.x - side1.x * side2.z;
    c.z = side1.x * side2.y - side1.y * side2.x;

    /*
     * Reverse the direction of the polygon's normal because the cross
     * product actually produces a vector which points away from the visible
     * side
     */

    c.x = -c.x;
    c.y = -c.y;
    c.z = -c.z;

    view.x = b->node->vertex[0].x - x;
    view.y = b->node->vertex[0].y - y;
    view.z = b->node->vertex[0].z - z;

    dot = Vec3D_RawDotProduct(&c, &view);
    return (dot <= 0) ? 1 : 0;
}
//#endif

static void DoPoly(BSP *b)
{
    if (b == NULL)
        return;

    /*
     * Traverse the tree properly (back to front)
     */

    if (PolyIsFacingUs(b))
    {
        DoPoly(b->back);    // if poly is facing us, do back polys first
        DrawPoly(b);
        DoPoly(b->front);
    }
    else
    {
        DoPoly(b->front);
        DoPoly(b->back);
    }
}

static void RenderWorld()
{
    DoPoly(bsp_tree);
}


/*****************************************************************************
 Video Code
*****************************************************************************/

static void Blit()
{
    UINT32  i, j, vd, vs;

    if (lfb.pitch == lfb.xres)
    {
        _farsetsel(lfb.selector);
        for (i = 0; i < lfb.xres * lfb.yres; i += 4)
            _farnspokel(i, *((UINT32 *) &vidbuf[i]));
        return;
    }
    _farsetsel(lfb.selector);
    for (i = vd = vs = 0; i < lfb.yres; i++)
    {
        for (j = 0; j < lfb.xres / 4; j ++, vd += 4, vs += 4)
//            _farpokel(lfb.selector, vd, *((unsigned int *) (vidbuf + vs)));
            _farnspokel(vd, *((unsigned int *) (vidbuf + vs)));
        vd += lfb.pitch - lfb.xres;
    }
}

static void TextMode()
{
    __dpmi_regs r;

    r.x.ax = 0x0003;
    __dpmi_int(0x10, &r);
}

static INT32 VideoMode(INT32 xres, INT32 yres)
{
    UINT16  mode;

    if (VESALib_CheckVESA())
    {
        fprintf(stderr, "VESA 2.0 is not available!\n");
        return 1;
    }

    VESALib_CreateModeList();

    if (VESALib_FindMode(xres, yres, 8, &mode))
    {
        fprintf(stderr, "Video mode not supported: %dx%dx8\n", xres, yres);
        return 1;
    }

    if ((vidbuf = calloc(xres * yres, sizeof(UINT8))) == NULL)
    {
        fprintf(stderr, "Out of memory!\n");
        return 1;
    }

    if (VESALib_SetMode(mode))
    {
        fprintf(stderr, "Unable to set video mode: %dx%dx8\n", xres, yres);
        return 1;
    }

    if (VESALib_MapLinearFrameBuffer())
    {
        TextMode();
        fprintf(stderr, "Unable to map linear frame buffer!\n");
        return 1;
    }

    VESALib_GetLinearFrameBufferInfo(&lfb);
    return 0;
}


/*****************************************************************************
 BSP Management
*****************************************************************************/

static void KillBSPTree(BSP *b)
{
    if (b == NULL)
        return;

    KillBSPTree(b->front);
    KillBSPTree(b->back);

    free(b);
}

static void KillNodeList()
{
    NODE    *next;

    while (node_list != NULL)
    {
        next = node_list->next;
        free(node_list->vertex);
        free(node_list);
        node_list = next;
    }
}

static NODE *FindNode(INT id)
{
    NODE    *n;

    if (!id) return NULL;
    for (n = node_list; n != NULL; n = n->next)
    {
        if (n->id == id)
            break;
    }

    if (n == NULL && id != 0)
        printf("error: node not found (%d)\n", id);
    return n;   // this will return NULL if there was no match
}

static void MakeBSPNode(BSP **b, NODE *n)
{
    if (n == NULL)
        return;

    if ((*b = malloc(sizeof(BSP))) == NULL)
    {
        KillBSPTree(bsp_tree);
        KillNodeList();
        fprintf(stderr, "r: out of memory building BSP tree\n");
        exit(1);
    }

    (*b)->node = n;
    (*b)->front = NULL;
    (*b)->back = NULL;

    MakeBSPNode(&((*b)->front), FindNode(n->front_id));
    MakeBSPNode(&((*b)->back), FindNode(n->back_id));
}

static void BuildBSPTree()
{
    MakeBSPNode(&bsp_tree, node_list);
}

static void NextColor(UINT8 *color)
{
//    *color = rand() % 0xf;
    (*color)++;
    (*color) &= 0xf;    // first 16 colors
    if (*color == 0 || *color == 8) // but not these!
        (*color)++;
}

static BOOL LoadBSPNodes(char *file)
{
    FILE    *fp;
    NODE    *n, *nsearch;
    INT     i, id, front_id, back_id, num_vertices;
    UINT8   color;

    srand(time(NULL));

    /*
     * Read in all of the nodes and then build a BSP tree
     *
     * NOTE: First node must be the root
     */

    if ((fp = fopen(file, "r")) == NULL)
    {
        fprintf(stderr, "r: unable to open %s\n", file);
        return 1;
    }

    color = 1;
    while (1)
    {

        fscanf(fp, "%d,%d,%d:%d", &id, &front_id, &back_id, &num_vertices);
        if (!num_vertices)
            break;

        if ((n = malloc(sizeof(NODE))) == NULL)
        {
            fclose(fp);
            KillNodeList();
            fprintf(stderr, "r: out of memory!\n");
            return 1;
        }

        if ((n->vertex = malloc(num_vertices * sizeof(VERTEX3D))) == NULL)
        {
            fclose(fp);
            free(n);
            KillNodeList();
            fprintf(stderr, "r: out of memory!\n");
            return 1;
        }

        for (i = 0; i < num_vertices; i++)
            fscanf(fp, "%f,%f,%f", &(n->vertex[i].x), &(n->vertex[i].y), &(n->vertex[i].z));

        n->next = NULL; // link into end of list
        n->num_vertices = num_vertices;
        n->id = id;
        n->front_id = front_id;
        n->back_id = back_id;
        n->color = color;
        NextColor(&color);

        if (node_list == NULL)  // first node in list?
            node_list = n;      // attach directly to list
        else
        {
            for (nsearch = node_list; nsearch->next != NULL; nsearch = nsearch->next)
                ;
            nsearch->next = n;
        }
    }

    fclose(fp);
    return 0;
}

static void OutputBSPTree(BSP *b)
{
    INT i;

    if (b == NULL)
        return;

    printf("%X,%X,%X\n", (UINT) b, (UINT) b->front, (UINT) b->back);
    for (i = 0; i < b->node->num_vertices; i++)
        printf("%f,%f,%f\n", b->node->vertex[i].x, b->node->vertex[i].y, b->node->vertex[i].z);

    OutputBSPTree(b->front);
    OutputBSPTree(b->back);
}


int main(int argc, char **argv)
{
    NODE    *n;
    INT     quit, i, j;
    UINT    move = 0;   // move up | move down | roll left | roll right | look up | look down | left | right | forward | back
    UINT8   k;

    if (argc <= 1)
    {
        puts("r: no input file, nothing to do");
        exit(0);
    }

    node_list = NULL;
    bsp_tree = NULL;

    if (LoadBSPNodes(argv[1]))
        exit(1);

    i = 3;          // find the biggest number of vertices
    n = node_list;
    while (n != NULL)
    {
        if (n->num_vertices > i)
            i = n->num_vertices;
        n = n->next;
    }
    j = ((vlist_a = malloc((i * 2) * sizeof(VERTEX3D))) == NULL) ? 1 : 0;
    j |= ((vlist_b = malloc((i * 2) * sizeof(VERTEX3D))) == NULL) ? 1 : 0;
    j |= ((vlist_2d = malloc((i * 2) * sizeof(VERTEX2D))) == NULL) ? 1 : 0;
    if (j)
    {
        fprintf(stderr, "r: out of memory!\n");
        if (vlist_a != NULL)
            free(vlist_a);
        if (vlist_b != NULL)
            free(vlist_b);
        if (vlist_2d != NULL)
            free(vlist_2d);
        KillNodeList();
        exit(1);
    }

    BuildBSPTree();
//    OutputBSPTree(bsp_tree);
//    exit(0);

    /*
     * Set video mode
     */

    if (argc <= 3)
    {
        xres = 640;
        yres = 480;
    }
    else
    {
        xres = atoi(argv[2]);
        yres = atoi(argv[3]);
    }

    dphys = (xres / 2) / tan(DEG2RAD(FOV / 2));
    fov_y = atan((yres / 2) / dphys) * 2;
    fov_y = RAD2DEG(fov_y);

    if (KeybInit())
    {
        free(vlist_a);
        free(vlist_b);
        free(vlist_2d);
        KillNodeList();
        KillBSPTree(bsp_tree);
        exit(1);
    }

    if (VideoMode(xres, yres))
    {
        free(vlist_a);
        free(vlist_b);
        free(vlist_2d);
        KillNodeList();
        KillBSPTree(bsp_tree);
        exit(1);
    } 

    /*
     * Let there be BSP!
     */

    quit = 0;
    while (!quit)
    {
        memset(vidbuf, 0, xres * yres); // erase

        k = KeybGetKey();
        switch (k)
        {
        case KEYB_KEYP_P:
            RenderWorld();
            ScreenShot(vidbuf, xres, yres);
            break;
        case KEYB_KEYP_ESC:
            quit = 1;
            break;
        case KEYB_KEYP_UP:
            move |= 2;
            move &= ~1;
            break;
        case KEYB_KEYR_UP:
            move &= ~2;
            break;
        case KEYB_KEYP_DOWN:
            move |= 1;
            move &= ~2;
            break;
        case KEYB_KEYR_DOWN:
            move &= ~1;
            break;
        case KEYB_KEYP_LEFT:
            move |= 8;
            move &= ~4;
            break;
        case KEYB_KEYR_LEFT:
            move &= ~8;
            break;
        case KEYB_KEYP_RIGHT:
            move |= 4;
            move &= ~8;
            break;
        case KEYB_KEYR_RIGHT:
            move &= ~4;
            break;
        case KEYB_KEYP_A:
            move |= 0x20;
            move &= ~0x10;
            break;
        case KEYB_KEYR_A:
            move &= ~0x20;
            break;
        case KEYB_KEYP_Z:
            move |= 0x10;
            move &= ~0x20;
            break;
        case KEYB_KEYR_Z:
            move &= ~0x10;
            break;
        case KEYB_KEYP_S:
            move |= 0x200;
            move &= ~0x100;
            break;
        case KEYB_KEYR_S:
            move &= ~0x200;
            break;
        case KEYB_KEYP_X:
            move |= 0x100;
            move &= ~0x200;
            break;
        case KEYB_KEYR_X:
            move &= ~0x100;
            break;
        case KEYB_KEYP_COMMA:
            move |= 0x80;
            move &= ~0x40;
            break;
        case KEYB_KEYR_COMMA:
            move &= ~0x80;
            break;
        case KEYB_KEYP_DOT:
            move |= 0x40;
            move &= ~0x80;
            break;
        case KEYB_KEYR_DOT:
            move &= ~0x40;
            break;
        case KEYB_KEYP_NUMPAD5:
            pitch = roll = 0.0;
            break;
        default:
            break;
        }

        if (move & 2)       // Forward
        {
            x += sin(DEG2RAD(-yaw))*5;
            z += cos(DEG2RAD(yaw))*5;
        }
        if (move & 1)       // Back
        {
            x -= sin(DEG2RAD(-yaw))*5;
            z -= cos(DEG2RAD(yaw))*5;
        }
        if (move & 8)       // Left
            yaw += 1;
        if (move & 4)       // Right
            yaw -= 1;
        if (move & 0x200)   // Up
            y -= 5;
        if (move & 0x100)   // Down
            y += 5;
        if (move & 0x80)    // Roll Counter-clockwise
            roll--;
        if (move & 0x40)    // Roll Clockwise
            roll++;
        if (move & 0x20)    // Look Up
            pitch--;
        if (move & 0x10)    // Look Down
            pitch++;

        RenderWorld();

        while (inportb(0x3da) & 8)      ;
        while (!(inportb(0x3da) & 8))   ;
        Blit();
    }

    /*
     * Quit!
     */

    free(vlist_a);
    free(vlist_b);
    free(vlist_2d);
    TextMode();
    KeybDeinit();
    KillNodeList();
    KillBSPTree(bsp_tree);

    return 0;
}

