/*
 * vesalib.c
 *
 * VESALib Version 0.0: VESA 2.0 programming library for DJGPP
 * January 14 and November 28-29, 2001 by Bart Trzynadlowski
 *
 * Some VESA programming notes to spare the reader headaches if he/she is
 * attempting to write his/her own VESA code from scratch:
 *
 *      - When you get the VESA mode information, go through the list of mode
 *        numbers and store them locally before calling any other VESA
 *        functions, such as the "get mode info" function. Some cards, such
 *        as ATI's Rage128 apparently will overwrite the mode list or
 *        something if you use some other function calls.
 *      - Make sure that your VESA structures are correct. A tutorial I once
 *        saw on VESA had a structure set up which was missing a byte; this
 *        caused the linear frame buffer address to always be 0.
 */

#include <dpmi.h>
#include <go32.h>
#include <string.h>
#include <sys/farptr.h>
#include <sys/movedata.h>
#include "types.h"
#include "vesalib.h"


/*****************************************************************************
 Internal Data
*****************************************************************************/

static VESALIB_INFO     vesa_info;      // VESA information
static VESALIB_MODEINFO vesa_modeinfo;  // VESA mode information
static VESALIB_LFBINFO  lfb_info;       // linear frame buffer information
static VESALIB_MODES    modes;          // list of detected VESA modes
static INT32            current_mode;   // number of current mode


/*****************************************************************************
 VESALib Functions
*****************************************************************************/

/*
 * INT32 VESALib_CheckVESA();
 *
 * Checks for VESA 2.0. Returns VESALIB_ERROR_NOVESA if VESA 2.0 is not
 * present.
 */

INT32 VESALib_CheckVESA()
{
    if (VESALib_GetVESAInfo(&vesa_info))
        return VESALIB_ERROR_NOVESA;

    if (strncmp("VESA", vesa_info.VESASignature, 4) || vesa_info.VESAVersion < 0x200)
        return VESALIB_ERROR_NOVESA;    // no VESA 2.0!

    return 0;
}

/*
 * INT32 VESALib_GetVESAInfo(VESALIB_INFO *info);
 *
 * Get VESA information. Returns VESALIB_ERROR_NOVESA if an error occured.
 * This function does not check the data returned like VESALib_CheckVESA()
 * does.
 */

INT32 VESALib_GetVESAInfo(VESALIB_INFO *info)
{
    __dpmi_regs r;

    dosmemput("VBE2", 4, __tb);

    r.x.ax = 0x4f00;
    r.x.di = __tb & 0xf;
    r.x.es = (__tb >> 4) & 0xffff;
    __dpmi_int(0x10, &r);

    dosmemget(__tb, sizeof(VESALIB_INFO), info);

    if (r.h.ah)
        return VESALIB_ERROR_NOVESA;

    return 0;
}

/*
 * void VESALib_CreateModeList();
 *
 * Fills the mode list.
 */

void VESALib_CreateModeList()
{
    UINT32  mptr, i;

    modes.num_modes = 0;

    /*
     * Convert the video mode pointer from SEG:OFFS format into a 20-bit
     * linear address and catalog up to 256 mode numbers
     */

    mptr = ((vesa_info.VideoModePtr & 0xffff0000) >> 12) + (vesa_info.VideoModePtr & 0xffff);
    for (i = 0; i < 256; i++, mptr += 2)
    {
        if (_farpeekw(_dos_ds, mptr) == 0xffff)
            break;
        modes.mode_list[i] = _farpeekw(_dos_ds, mptr);
    }
    modes.num_modes = i;
}

/*
 * VESALIB_MODES *VESALib_GetModeList();
 *
 * Returns a pointer to the mode list structure. This is only valid after
 * VESALib_CreateModeList() has been called.
 */

VESALIB_MODES *VESALib_GetModeList()
{
    return &modes;
}

/*
 * INT32 VESALib_GetModeInfo(UINT16 mode, VESALIB_MODEINFO *mode_info);
 *
 * Gets information on the VESA mode and stores it in mode_info. Returns
 * VESALIB_ERROR_GETMODEINFOFAILED if something went wrong.
 */

INT32 VESALib_GetModeInfo(UINT16 mode, VESALIB_MODEINFO *mode_info)
{
    __dpmi_regs r;

    /*
     * Get mode information
     */

    r.x.ax = 0x4f01;
    r.x.cx = mode;
    r.x.di = __tb & 0xf;
    r.x.es = (__tb >> 4) & 0xffff;
    __dpmi_int(0x10, &r);
    
    if (r.h.ah)                                 // did function succeed?
        return VESALIB_ERROR_GETMODEINFOFAILED; // nope!

    dosmemget(__tb, sizeof(VESALIB_MODEINFO), mode_info);

    return 0;
}

/*
 * INT32 VESALib_FindMode(UINT16 xres, UINT16 yres, UINT8 bpp, UINT16 *mode);
 *
 * Finds the mode with the indicated resolution and bits per pixel and returns
 * its mode number in the word pointed to by mode. In addition to the
 * requested attributes, the mode must have 1 plane, must be a color graphics
 * mode, must be supported in hardware, and must support a linear frame
 * buffer. If a matching mode is not found, VESALIB_ERROR_MODENOTAVAILABLE is
 * returned.
 */

INT32 VESALib_FindMode(UINT16 xres, UINT16 yres, UINT8 bpp, UINT16 *mode)
{
    INT32   i, ret_val;

    /*
     * Scan through the list and find a mode we can use
     */

    for (i = 0, ret_val = VESALIB_ERROR_MODENOTAVAILABLE; i < modes.num_modes; i++)
    {
        if (VESALib_GetModeInfo(modes.mode_list[i], &vesa_modeinfo))
            continue;   // ignore errors, we can just skip over bad modes

        /*
         * Check to make sure:
         * - X resolution is as requested
         * - Y resolution is as requested
         * - BPP is as requested
         * - Number of planes is 1
         * - The mode is supported in hardware, color, and a graphics mode
         * - A linear frame buffer is supported
         */

        if (vesa_modeinfo.XResolution == xres && vesa_modeinfo.YResolution == yres && vesa_modeinfo.BitsPerPixel == bpp && vesa_modeinfo.NumberOfPlanes == 1 && (vesa_modeinfo.ModeAttributes & 0x19) && (vesa_modeinfo.ModeAttributes & 0x80))
        {
            *mode = modes.mode_list[i];
            ret_val = 0;
            break;
        }
    }

    return ret_val;
}

/*
 * INT32 VESALib_SetMode(UINT16 mode);
 *
 * Sets the requested mode and returns VESALIB_ERROR_MODESETFAILED if it
 * failed to set the mode or a VESALib_GetModeInfo() error code if it failed
 * that function -- either way, the mode was not set. Information on the
 * linear frame buffer is obtained, but not completely. Wait until after you
 * successfully call VESALib_MapLinearFrameBuffer() to read out the frame
 * buffer information.
 */

INT32 VESALib_SetMode(UINT16 mode)
{
    __dpmi_regs r;
    INT32       i;

    /*
     * Get mode information
     */

    if ((i = VESALib_GetModeInfo(mode, &vesa_modeinfo)))
        return i;

    /*
     * Try setting the mode
     */

    r.x.ax = 0x4f02;
    r.x.bx = mode | 0x4000; // linear frame buffer (clear memory, too)
    __dpmi_int(0x10, &r);
    if (r.x.ax != 0x004f)
        return VESALIB_ERROR_MODESETFAILED;

    current_mode = mode;    // successful, so indicate we're in this mode

    /*
     * Get frame buffer information
     */

    lfb_info.pitch = vesa_modeinfo.BytesPerScanLine;
    lfb_info.xres = vesa_modeinfo.XResolution;
    lfb_info.yres = vesa_modeinfo.YResolution;
    lfb_info.bpp = vesa_modeinfo.BitsPerPixel;
    lfb_info.red_mask_size = vesa_modeinfo.RedMaskSize;
    lfb_info.red_field_pos = vesa_modeinfo.RedFieldPosition;
    lfb_info.green_mask_size = vesa_modeinfo.GreenMaskSize;
    lfb_info.green_field_pos = vesa_modeinfo.GreenFieldPosition;
    lfb_info.blue_mask_size = vesa_modeinfo.BlueMaskSize;
    lfb_info.blue_field_pos = vesa_modeinfo.BlueFieldPosition;

    return 0;
}

/*
 * UINT16 VESALib_GetCurrentMode();
 *
 * Returns the current mode number.
 */

UINT16 VESALib_GetCurrentMode()
{
    return current_mode;
}

/*
 * INT32 VESALib_MapLinearFrameBuffer();
 *
 * Maps the linear frame buffer for the currently set video mode to a selector
 * we can use. Returns either an error code from VESALib_GetModeInfo() if that
 * function failed or VESALIB_ERROR_LFBMAPPINGFAILED if the frame buffer could
 * not be mapped. This must be called after VESALib_SetMode().
 */

INT32 VESALib_MapLinearFrameBuffer()
{
    __dpmi_meminfo  info;
    INT32           i;

    if ((i = VESALib_GetModeInfo(current_mode, &vesa_modeinfo)))
        return i;

    info.size = vesa_modeinfo.BytesPerScanLine * vesa_modeinfo.YResolution;//vesa_info.TotalMemory * 64 * 1024;
    info.address = vesa_modeinfo.PhysBasePtr;
    if (__dpmi_physical_address_mapping(&info))
        return VESALIB_ERROR_LFBMAPPINGFAILED;

    lfb_info.selector = __dpmi_allocate_ldt_descriptors(1);
    __dpmi_set_segment_base_address(lfb_info.selector, info.address);
    __dpmi_set_segment_limit(lfb_info.selector, info.size - 1);
#if 0   // theoretically, you can add nearptr support with something like this
        // commented out code
    lfb_offs = (UINT8 *) (vesa_modeinfo.PhysBasePtr + __djgpp_conventional_base);
#endif

    return 0;
}

/*
 * void VESALib_GetLinearFrameBufferInfo(VESALIB_LFBINFO *info);
 *
 * Copies the information on the current linear frame buffer to the specified
 * location. This should be called after VESALib_MapLinearFrameBuffer().
 */

void VESALib_GetLinearFrameBufferInfo(VESALIB_LFBINFO *info)
{
    memcpy(info, &lfb_info, sizeof(VESALIB_LFBINFO));
}
